#
# Copyright (C) 2006 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

__all__ = ['BrowserApplet']

PROGNAME="setroubleshoot"
import gettext
gettext.bindtextdomain(PROGNAME, "/usr/share/locale")
gettext.textdomain(PROGNAME)
try:
    gettext.install(PROGNAME, "/usr/share/locale", 1)
except IOError:
    import __builtin__
    __builtin__.__dict__['_'] = unicode

# This import must come before importing gtk to silence warnings
from setroubleshoot.gui_utils import *

import os
import sys
import pygtk
pygtk.require("2.0")
import gtk, gtk.glade
import pango
import re
import threading
import traceback

import gobject 
import gnome.ui    

from setroubleshoot.analyze import *
from setroubleshoot.config import cfg
from setroubleshoot.errcode import *
from setroubleshoot.signature import *
from setroubleshoot.util import *
from setroubleshoot.log import *
from setroubleshoot.rpc import *
from setroubleshoot.rpc_interfaces import *
import gtkhtml2

#------------------------------------------------------------------------------

NUM_COLUMNS = 6
(PYOBJECT_COLUMN, FILTER_TYPE_COLUMN, DATE_COLUMN, COUNT_COLUMN, CATEGORY_COLUMN,
 SUMMARY_COLUMN) = range(NUM_COLUMNS)

#------------------------------------------------------------------------------

class LogfileThread(threading.Thread,
                    RpcManage,
                    SETroubleshootDatabaseNotifyInterface,
                    gobject.GObject):

    __gsignals__ = {
        'database_bind': 
        (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT)),
        'signatures_updated': 
        (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT)),
        }


    def __init__(self, browser, logfile_path):
        # parent class constructors
        threading.Thread.__init__(self)
        gobject.GObject.__init__(self)
        RpcManage.__init__(self)
        self.connect_rpc_interface('SETroubleshootDatabaseNotify', self)
        self.browser = browser
        self.logfile_path = logfile_path
        self.analyzer = LogfileAnalyzer(self.logfile_path, notify=self.browser)
        self.database = self.analyzer.database_rpc

    def run(self):
        self.process_logfile()

    def progress_callback(self, progress):
        self.browser.statusbar_widget.progress.set_fraction(progress)

    def process_logfile(self):
        if debug:
            log_gui.debug("on_open_logfile: self.logfile_path=%s", self.logfile_path)

        self.browser.signal_connection_database_bind.connect(self, self.browser.on_logfile_database_bind)
        self.emit('database_bind', self.database, self.analyzer.database.properties)
        self.browser.update_visit('logfile')

        self.browser.set_status_message('opening %s' % self.logfile_path)
        self.browser.statusbar_widget.progress.set_fraction(0.0)
        self.browser.statusbar_widget.progress.set_text(self.browser.logfile_database_properties.friendly_name)
        # FIXME: instead of load_data() it should clear the model, set initial selection
        self.browser.load_data()
        self.analyzer.analyze_logfile(self.progress_callback)

    def signatures_updated(self, type, item):
        if debug:
            log_rpc.debug('signatures_updated() LogfileThread: type=%s item=%s', type, item)
        self.emit('signatures_updated', type, item)
        
gobject.type_register(LogfileThread)

#------------------------------------------------------------------------------

class BrowserStatusBar(gtk.HBox):
    # These should sum to 1.0
    visit_msg_proportion  = 0.2
    status_msg_proportion = 0.6
    progress_proportion   = 0.2

    def __init__(self):
        gtk.HBox.__init__(self)
        self.connect('size-allocate', self.on_size_allocate)

        self.connect_icon = gtk.Image()
        self.connect_icon.set_from_stock(gtk.STOCK_DISCONNECT, gtk.ICON_SIZE_SMALL_TOOLBAR)
        self.connect_icon.show()
    
        self.connect_icon_frame = gtk.Frame()
        self.connect_icon_frame.add(self.connect_icon)
        self.pack_start(self.connect_icon_frame, False, False, 0)
        self.connect_icon_frame.show()
	
        self.visit_msg = gtk.Label("visiting")
        self.visit_msg.set_ellipsize(pango.ELLIPSIZE_END)
        self.visit_msg.set_alignment(0.0, 0.5)
        self.visit_msg_frame = gtk.Frame()
        self.visit_msg_frame.add(self.visit_msg)
        self.pack_start(self.visit_msg_frame, False, False, 0)
        self.visit_msg_frame.show()
        self.visit_msg.show()
	
        self.status_msg = gtk.Label("status")
        self.status_msg.set_ellipsize(pango.ELLIPSIZE_END)
        self.status_msg.set_alignment(0.0, 0.5)
        self.status_msg_frame = gtk.Frame()
        self.status_msg_frame.add(self.status_msg)
        self.pack_start(self.status_msg_frame, False, False, 0)
        self.status_msg_frame.show()
        self.status_msg.show()
	
        self.progress = gtk.ProgressBar()
        self.progress.set_ellipsize(pango.ELLIPSIZE_END)
        self.progress_frame = gtk.Frame()
        self.progress_frame.add(self.progress)
        self.pack_start(self.progress_frame, False, False, 0)
        self.progress_frame.show()
        self.progress.show()
	
        self.show()
        
    def on_size_allocate(self, widget, allocation):
        max_x = allocation.x + allocation.width
        
        child_rect = gtk.gdk.Rectangle(allocation.x, allocation.y,
                                 allocation.width, allocation.height)

        # Compute size for connection icon
        icon_width, icon_height = self.connect_icon_frame.size_request()

        # Total space available for fixed sized items
        fixed_width_total = icon_width
        fixed_width_total = min(allocation.width, fixed_width_total)
        fixed_width_remaining = fixed_width_total

        # Total space available remaining for scaled items
        scaled_width_total = allocation.width - fixed_width_total
        scaled_width_remaining = scaled_width_total

        # Move left to right laying out items:
        # [connection icon][visit msg][status msg][progress bar]

        # Connection Icon
        child_rect.width = min(fixed_width_total, icon_width)
        self.connect_icon_frame.size_allocate(child_rect)
        child_rect.x += child_rect.width
        fixed_width_remaining = fixed_width_remaining - child_rect.width

        # Visit Message
        child_rect.width = int(scaled_width_total * self.visit_msg_proportion)
        if (child_rect.x + child_rect.width) > max_x:
            child_rect.width = max_x - child_rect.x
        self.visit_msg_frame.size_allocate(child_rect)
        scaled_width_remaining = scaled_width_remaining - child_rect.width
        child_rect.x += child_rect.width

        # Status Message
        child_rect.width = int(scaled_width_total * self.status_msg_proportion)
        if (child_rect.x + child_rect.width) > max_x:
            child_rect.width = max_x - child_rect.x
        self.status_msg_frame.size_allocate(child_rect)
        scaled_width_remaining = scaled_width_remaining - child_rect.width
        child_rect.x += child_rect.width

        # Progress Bar
        # Note, we don't compute the proportionate space for the last scaled item
        # because of potential rounding errors, we just use whats left
        child_rect.width = scaled_width_remaining
        if child_rect.x + child_rect.width > max_x:
            child_rect.width = max(0, max_x - child_rect.x)
        self.progress_frame.size_allocate(child_rect)

        return True

#------------------------------------------------------------------------------

class SignalConnection(object):
    def __init__(self, signal, emitter=None, id=None):
        self.signal = signal
        self.emitter = emitter
        self.id = id

    def connect(self, emitter, handler):
        self.disconnect()
        self.emitter = emitter
        if self.emitter is not None:
            self.id = self.emitter.connect(self.signal, handler)
        

    def disconnect(self):
        if self.emitter is not None and self.id is not None:
            self.emitter.disconnect(self.id)
        self.emitter = None
        self.id = None

#------------------------------------------------------------------------------

class BrowserApplet(Applet):

    # --- Initialization ---

    def __init__(self, username, server):
        self.username = username
        self.server = server
        self.cur_siginfo = None
        self.alert_display = None
        self.window_delete_hides = True
        self.load_in_progress = False
        self.audit_database = None
        self.audit_database_properties = None
        self.logfile_database = None
        self.logfile_database_properties = None
        self.visit_target = None
        self.logfile_thread = None
        self.hide_deleted = False
        self.default_save_folder = get_user_home_dir()
        self.default_save_filename = 'selinux_alert.txt'
        self.clipboard = gtk.Clipboard()
        self.print_settings = None
        self.last_selected_row = 0

        # How long an alert must be displayed before it's marked as having been seen
        self.seen_after_displayed_seconds = 3

        program_name = _('setroubleshoot browser')
        program_version = '1.0'
        gnome.init(program_name, program_version)
        self.window_name = _("SETroubleshoot Browser")
        gtk.glade.set_custom_handler(self.glade_custom_handler)
        glade_file = cfg.get('alert','browser_glade_file')
        self.wTree = gtk.glade.XML(glade_file)
        self.init_widgets()

        window_state = os.getenv('SEALERT_WINDOW_STATE')
        if debug:
            log_gui.debug("read SEALERT_WINDOW_STATE from environment (%s)", window_state)
        if window_state is None:
            window_state = 'hidden'
        self.window_state = parse_window_state(window_state)
        self.update_window_state(self.window_state)

        window_geometry = os.getenv('SEALERT_WINDOW_GEOMETRY')
        if debug:
            log_gui.debug("read SEALERT_WINDOW_GEOMETRY from environment (%s)", window_geometry)
        if window_geometry is not None:
            self.set_geometry(window_geometry)

        self.update_connection_state(server.get_connection_state())

        self.signal_connection_connection_state_change  = SignalConnection('connection_state_change')
        self.signal_connection_connection_pending_retry = SignalConnection('connection_pending_retry')
        self.signal_connection_signatures_updated       = SignalConnection('signatures_updated')
        self.signal_connection_database_bind            = SignalConnection('database_bind')

        self.signal_connection_database_bind.connect(self.server, self.on_audit_database_bind)

        self.update_visit('audit')
        
        dic = {'on_alert_list_button_press'    : self.on_alert_list_button_press,
               'on_browser_delete_event'       : self.on_close,
               'on_open_logfile_activate'      : self.on_open_logfile,
               'on_close_activate'             : self.on_close,
               'on_copy_activate'              : self.on_copy,
               'on_copy_alert_activate'        : self.on_copy_alert,
               'on_print_activate'             : self.on_print,
               'on_delete_activate'            : self.on_delete,
               'on_undelete_activate'          : self.on_undelete,
               'on_expunge_activate'           : self.on_expunge,
               'on_select_all_activate'        : self.on_select_all,
               'on_select_none_activate'       : self.on_select_none,
               'on_save_as_activate'           : self.on_save_as,
               'on_view_audit_activate'        : self.on_view_audit,
               'on_view_logfile_activate'      : self.on_view_logfile,
               'on_report_bug_activate'        : self.on_report_bug,
               'on_about_activate'             : self.on_about,
               } 
        self.wTree.signal_autoconnect(dic) 

    def glade_custom_handler (self, glade, function_name, widget_name, str1, str2, int1 , int2):
        handler = getattr(self, function_name)
        if handler is None:
            return None
	return handler(widget_name, str1, str2, int1, int2)

    def create_statusbar(self, widget_name, str1, str2, int1 , int2):
        widget = BrowserStatusBar()
        return widget

    def on_realize(self, widget):
        # position the divider between the top and bottom pane
        vpane_rect = self.vpane_widget.get_allocation()
        self.vpane_widget.set_position(vpane_rect.height/4)

    def get_geometry(self):
        width, height = self.browser_widget.get_size()
        xoffset, yoffset = self.browser_widget.get_position()
        geometry = '%dx%d+%d+%d' % (width, height, xoffset, yoffset)
        if debug:
            log_gui.debug("get_geometry() %s", geometry)
        return geometry

    def set_geometry(self, geometry):
        screen_width = gtk.gdk.screen_width()
        screen_height = gtk.gdk.screen_height()
        match = re.search("(\d+)x(\d+)", geometry)
        if match:
            width  = int(match.group(1))
            height = int(match.group(2))
            #print "width=%d height=%d" % (width, height)
            self.browser_widget.resize(width, height)

        match = re.search("([+-]\d+)([+-]\d+)", geometry)
        if match:
            xoffset = int(match.group(1))
            yoffset = int(match.group(2))
            #print "xoffset=%d yoffset=%d" % (xoffset, yoffset)
            if xoffset >= 0:
                x = xoffset
            else:
                x = screen_width() + xoffset

            if yoffset >= 0:
                y = yoffset
            else:
                y = screen_height()+ yoffset
            
            x = max(x, 0)
            y = max(y, 0)
            # force window to be visible
            if x >= screen_width - 20:
                x = 0
            if y >= screen_height - 20:
                y = 0
            self.browser_widget.move(x, y)

    def window_state_event_cb(self, window, event):
        if debug:
            log_gui.debug("window state event: %s", window_state_to_string(event.new_window_state))
        self.window_state = event.new_window_state

    def configure_event_cb(self, window, event):
        if debug:
            log_gui.debug("configure event: x=%d y=%d width=%d height=%d",
                          event.x, event.y, event.width, event.height)

    def init_widgets(self):
        self.get_widget('browser')
        self.browser_widget.connect_after("realize", self.on_realize)
        self.browser_widget.connect('window-state-event', self.window_state_event_cb)
        self.browser_widget.connect('configure-event', self.configure_event_cb)

        self.get_widget("vpane")
        self.get_widget("statusbar")

        self.init_list_view()

        # description
        self.get_widget("alert_detail_scrolled_window")
        view, doc = self.create_htmlview(self.alert_detail_scrolled_window_widget)
        doc.connect("link-clicked", self.link_clicked)
        self.description_doc = doc
        self.description_view = view

        self.get_widget("hide_deleted")
        self.hide_deleted_widget.set_active(self.hide_deleted)
        self.hide_deleted_widget.connect('toggled', self.on_hide_deleted_toggle)
        

        self.get_widget('alert_list_right_click_menu')

    def init_list_view(self):
        self.alert_list_widget = self.get_widget('alert_list')

        # create the base model
        self.alert_list_model = gtk.ListStore(gobject.TYPE_PYOBJECT,# siginfo
                                              gobject.TYPE_BOOLEAN, # filter
                                              gobject.TYPE_PYOBJECT,# date
                                              gobject.TYPE_INT,     # count
                                              gobject.TYPE_STRING,  # category
                                              gobject.TYPE_STRING)  # summary

        # create an intermediate model to filter rows with
        self.alert_list_model_filter = self.alert_list_model.filter_new()
        self.alert_list_model_filter.set_visible_func(self.visible_row_filter)

        # create an intermediate model to sort the filtered rows with
        self.alert_list_model_sort = gtk.TreeModelSort(self.alert_list_model_filter)

        # set the TreeView's model to the sortable model,
        # which is backed by the filtered model, which is backed by the base list store model
        self.alert_list_widget.set_model(self.alert_list_model_sort)

        #
        # Now create each column and intitialize it's cell renderers, etc.
        #

        # --- siginfo ---
        tv_column = gtk.TreeViewColumn()
        tv_column.set_visible(False)
        self.alert_list_widget.append_column(tv_column)

        # --- filter ---
        cell = gtk.CellRendererToggle()
        cell.set_property('activatable', True)
        cell.connect( 'toggled', self.on_filter_toggle, self.alert_list_model_sort)
        tv_column = gtk.TreeViewColumn('Filter', cell)
        tv_column.set_cell_data_func(cell, self.filter_type_cell_data, FILTER_TYPE_COLUMN)
        tv_column.set_sort_column_id(FILTER_TYPE_COLUMN)
        self.alert_list_widget.append_column(tv_column)

        # --- date ----
        cell = gtk.CellRendererText()
        tv_column = gtk.TreeViewColumn('Date', cell)
        tv_column.set_cell_data_func(cell, self.date_cell_data, DATE_COLUMN)
        tv_column.set_sort_column_id(DATE_COLUMN)
        self.alert_list_model_sort.set_sort_func(DATE_COLUMN, self.sort_date_column, DATE_COLUMN)
        self.alert_list_widget.append_column(tv_column)

        # --- count ---
        cell = gtk.CellRendererText()
        tv_column = gtk.TreeViewColumn('Count', cell, text=COUNT_COLUMN)
        tv_column.set_cell_data_func(cell, self.count_cell_data, COUNT_COLUMN)
        tv_column.set_sort_column_id(COUNT_COLUMN)
        self.alert_list_widget.append_column(tv_column)

        # --- category ---
        cell = gtk.CellRendererText()
        tv_column = gtk.TreeViewColumn('Category', cell, text=CATEGORY_COLUMN)
        tv_column.set_cell_data_func(cell, self.text_cell_data, CATEGORY_COLUMN)
        tv_column.set_sort_column_id(CATEGORY_COLUMN)
        self.alert_list_widget.append_column(tv_column)

        # --- summary ---
        cell = gtk.CellRendererText()
        tv_column = gtk.TreeViewColumn('Summary', cell, text=SUMMARY_COLUMN)
        tv_column.set_cell_data_func(cell, self.text_cell_data, SUMMARY_COLUMN)
        tv_column.set_sort_column_id(SUMMARY_COLUMN)
        self.alert_list_widget.append_column(tv_column)

        #
        # Some final properties
        #

        # alternate row color for easy reading
        self.alert_list_widget.set_rules_hint(True)

        # Set up the selection objects
        self.alert_list_selection = self.alert_list_widget.get_selection()
        self.alert_list_selection.set_mode(gtk.SELECTION_MULTIPLE)
        self.alert_list_selection.connect('changed', self.on_selection_changed)

        # initially select first row
        self.alert_list_selection.select_path((0))
        model, row_paths = self.alert_list_selection.get_selected_rows()

        # set initial sort order
        self.sort_column_id = DATE_COLUMN
        self.sort_order = gtk.SORT_DESCENDING
        self.alert_list_model_sort.set_sort_column_id(self.sort_column_id, self.sort_order)

    def on_report_bug(self, widget):
        # FIXME - Should be specific to the alert
        os.spawnl(os.P_NOWAIT, "/usr/bin/htmlview","htmlview", "http://bugzilla.redhat.com/bugzilla/enter_bug.cgi")
        return True
    

    def link_clicked(self, arg1, arg2):
        os.spawnl(os.P_NOWAIT, "/usr/bin/htmlview","htmlview", arg2)
        
    # --- Row Handling ---

    def visible_row_filter(self, model, iter):
        visible = True

        siginfo = model.get_value(iter, PYOBJECT_COLUMN)
        if siginfo is not None:
            user_data = siginfo.get_user_data(self.username)

            if user_data.delete_flag and self.hide_deleted:
                visible = False

        return visible
        

    # --- Column Handling ---

    def sort_date_column(self, model, iter1, iter2, column_index):
        timestamp1 = model.get_value(iter1, column_index)
        timestamp2 = model.get_value(iter2, column_index)
        if timestamp1 is None or timestamp2 is None:
            return 0
        return timestamp1.cmp(timestamp2)

    def count_cell_data(self, column, cell, model, iter, column_index):
        siginfo = model.get_value(iter, PYOBJECT_COLUMN)
        text_attributes = self.get_text_attributes(siginfo)
        cell.set_property("attributes", text_attributes)

        # FIXME: this should be a global property, not set each time
        cell.set_property("xalign", 1.0)

        text = model.get_value(iter, column_index)
        cell.set_property('text', text) 
        return

    def date_cell_data(self, column, cell, model, iter, column_index):
        siginfo = model.get_value(iter, PYOBJECT_COLUMN)
        text_attributes = self.get_text_attributes(siginfo)
        cell.set_property("attributes", text_attributes)
        timestamp = model.get_value(iter, column_index)
        if timestamp is None:
            return
        cell.set_property('text', timestamp.format())
        return

    def text_cell_data(self, column, cell, model, iter, column_index):
        siginfo = model.get_value(iter, PYOBJECT_COLUMN)
        text_attributes = self.get_text_attributes(siginfo)
        cell.set_property("attributes", text_attributes)
        text = model.get_value(iter, column_index)
        cell.set_property('text', text) 
        return

    def filter_type_cell_data(self, column, cell, model, iter, column_index):
        siginfo = model.get_value(iter, PYOBJECT_COLUMN)
        user_data = siginfo.get_user_data(self.username)

        if user_data.filter.filter_type == FILTER_NEVER:
            cell.set_property('active', False)
        else:
            cell.set_property('active', True)
        return

    # --- Utilities ---

    def mark_seen(self, database, siginfo):
        if self.cur_siginfo == siginfo:
            user_data = siginfo.get_user_data(self.username)
            user_data.seen_flag = True
            self.alert_list_widget.queue_draw() # FIXME: can just this one row be updated instead of whole list?
            database.set_user_data(siginfo.sig, self.username, 'seen_flag', True)
        return False                    # False ==> do not call again, True ==> call again

    def mark_delete(self, database, siginfo, value=True):
        user_data = siginfo.get_user_data(self.username)
        user_data.delete_flag = value
        self.alert_list_widget.queue_draw() # FIXME: can just this one row be updated instead of whole list?
        database.set_user_data(siginfo.sig, self.username, 'delete_flag', value)
        self.alert_list_model_filter.refilter()
        return False

    def get_text_attributes(self, siginfo):
        user_data = siginfo.get_user_data(self.username)

        text_attributes = pango.AttrList()
        
        if not user_data.seen_flag:
            text_attributes.insert(pango.AttrWeight(pango.WEIGHT_HEAVY, 0, -1))

        if user_data.delete_flag:
            text_attributes.insert(pango.AttrStrikethrough(True, 0, 1000))

        return text_attributes
        

    def get_selected_row(self):
        model, row_paths = self.alert_list_selection.get_selected_rows()
        if len(row_paths) == 0:
            return None
        return row_paths[0]

    def reset_selection(self):
        n_rows = self.alert_list_model_sort.iter_n_children(None)
        if n_rows == 0:
            return
        new_row = min(self.last_selected_row, n_rows-1)
        self.alert_list_selection.select_path((new_row))

    def get_selected_siginfo(self):
        model, row_paths = self.alert_list_selection.get_selected_rows()
        if model is None or len(row_paths) == 0:
            return None
        path = row_paths[0]
        self.last_selected_row = path[0]
        iter = model.get_iter(path)
        siginfo = model.get_value(iter, PYOBJECT_COLUMN)
        return siginfo

    def get_selected_siginfos(self):
        model, row_paths = self.alert_list_selection.get_selected_rows()
        siginfos = []
        for path in row_paths:
            iter = model.get_iter(path)
            siginfo = model.get_value(iter, PYOBJECT_COLUMN)
            siginfos.append(siginfo)
        return siginfos

    def get_siginfos(self, criteria=None):
        siginfos = []
        iter = self.alert_list_model.get_iter_first()
        while iter:
            siginfo = self.alert_list_model.get_value(iter, PYOBJECT_COLUMN)
            siginfos.append(siginfo)
            iter = self.alert_list_model.iter_next(iter)
        return siginfos

            

    # --- View Management ---

    def show(self):
        self.browser_widget.show()

    def hide(self):
        self.browser_widget.hide()

    def update_window_state(self, state):
        if debug:
            log_gui.debug("set_window_state() %s:%s",
                          state, window_state_to_string(state))

        if state & gtk.gdk.WINDOW_STATE_ICONIFIED:
            self.browser_widget.iconify()
        else:
            self.browser_widget.deiconify()

        if state & gtk.gdk.WINDOW_STATE_MAXIMIZED:
            self.browser_widget.maximize()
        else:
            self.browser_widget.unmaximize()

        if state & gtk.gdk.WINDOW_STATE_STICKY:
            self.browser_widget.stick()

        if state & gtk.gdk.WINDOW_STATE_FULLSCREEN:
            self.browser_widget.fullscreen()
        else:
            self.browser_widget.unfullscreen()

        if state & gtk.gdk.WINDOW_STATE_ABOVE:
            self.browser_widget.set_keep_above(True)
        else:
            self.browser_widget.set_keep_above(False)

        if state & gtk.gdk.WINDOW_STATE_BELOW:
            self.browser_widget.set_keep_below(True)
        else:
            self.browser_widget.set_keep_below(False)

        if state & gtk.gdk.WINDOW_STATE_WITHDRAWN:
            self.browser_widget.hide()
        else:
            self.browser_widget.show()

    def get_window_state(self):
        return window_state_to_string(self.window_state)
    
    def __get_table_color(self):
        style = self.browser_widget.get_style()
        c = style.bg[gtk.STATE_NORMAL]
        color = "#%.2X%.2X%.2X" % (c.red / 256, c.green / 256, c.blue / 256)
        return color

    def update_alert_view(self):
        siginfo = self.get_selected_siginfo()

        if siginfo is None:
            self.clear_alert_view()
            self.reset_selection()
            return False
        self.cur_siginfo = siginfo
        self.description_doc.clear()
        self.description_doc.open_stream("text/html")
        html = (siginfo.format_html(self.__get_table_color()))
        if debug:
            #log_gui.debug(html)
            pass
        self.description_doc.write_stream(html)
        self.description_doc.close_stream()

        self.alert_detail_scrolled_window_widget.show_all()
        
        user_data = siginfo.get_user_data(self.username)
        if not user_data.seen_flag:
            gobject.timeout_add(1000*self.seen_after_displayed_seconds,
                                self.mark_seen, self.get_visit_database(), siginfo)
        return False                    # False ==> do not call again, True ==> call again

    def clear_alert_view(self):
        self.description_doc.clear()
        self.alert_detail_scrolled_window_widget.queue_draw()
        #self.alert_detail_scrolled_window_widget.show_all()

    def init_gui_data(self):
        self.alert_list_model.clear()
        count = 0
        for siginfo in self.sigs.signature_list:
            count += 1
            self.new_model_row(siginfo)
        # filter the new data
        self.alert_list_model_filter.refilter()
        self.alert_list_selection.select_path((0))

    def find_local_id_in_model(self, local_id, model=None):
        if model is None:
            model = self.alert_list_model
        iter = model.get_iter_first()
        while iter:
            row_siginfo = model.get_value(iter, PYOBJECT_COLUMN)
            if row_siginfo.local_id == local_id:
                return iter
            iter = model.iter_next(iter)

        return None

    def insert_siginfo_into_model(self, siginfo):
        iter = self.find_local_id_in_model(siginfo.local_id)
        if iter is None:
            if debug:
                log_gui.debug("insert_siginfo_into_model(): new")
            iter = self.new_model_row(siginfo)
        else:
            if debug:
                log_gui.debug("insert_siginfo_into_model(): replace")
            self.update_model_row(iter, siginfo)

    def update_model_row(self, iter, siginfo):
        count = siginfo.report_count
        date = siginfo.last_seen_date
        category = get_text_default(siginfo.category)
        summary = remove_linebreaks(siginfo.solution.summary)

        self.alert_list_model.set(iter, PYOBJECT_COLUMN, siginfo)
        self.alert_list_model.set(iter, COUNT_COLUMN,    count)
        self.alert_list_model.set(iter, DATE_COLUMN,     date)
        self.alert_list_model.set(iter, CATEGORY_COLUMN, category)
        self.alert_list_model.set(iter, SUMMARY_COLUMN,  summary)

    def new_model_row(self, siginfo):
        iter = self.alert_list_model.append()
        self.update_model_row(iter, siginfo)
        return iter

    def get_visit_database(self):
        database = None
        if self.visit_target == 'audit':
            database = self.audit_database
        elif self.visit_target == 'logfile':
            database = self.logfile_database
        return database

    def update_visit(self, target):
        if self.visit_target == target:
            #return
            pass

        self.visit_target = target

        self.clear_visit_message()
        if target == 'audit':
            self.signal_connection_connection_state_change.connect(self.server, self.on_connection_state_change)
            self.signal_connection_connection_pending_retry.connect(self.server, self.on_connection_pending_retry)
            self.signal_connection_signatures_updated.connect(self.server, self.on_signatures_updated)
            if self.audit_database_properties is not None:
                self.set_visit_message(self.audit_database_properties.friendly_name)
            else:
                self.set_visit_message(_('Unconnected Audit'))
        
        elif target == 'logfile':
            self.signal_connection_connection_state_change.connect(self.server, self.on_connection_state_change)
            self.signal_connection_connection_pending_retry.disconnect()
            if self.logfile_database_properties is not None:
                self.signal_connection_signatures_updated.connect(self.logfile_thread, self.on_signatures_updated)
                self.set_visit_message(self.logfile_database_properties.friendly_name)
            else:
                self.signal_connection_signatures_updated.disconnect()
                self.set_visit_message(_('Unspecified logfile'))

        self.load_data()

    # --- Data Handling ---

    def progress_pulse(self):
        if self.load_in_progress:
            self.statusbar_widget.progress.pulse()
            return True                 # call again
        else:
            return False                # do not call again

    def on_audit_database_bind(self, emitter, database, properties):
        self.audit_database = database
        self.audit_database_properties = properties
        if self.visit_target == 'audit':
            self.set_visit_message(self.audit_database_properties.friendly_name)
            self.load_data()

    def on_logfile_database_bind(self, emitter, database, properties):
        self.logfile_database = database
        self.logfile_database_properties = properties
        if self.visit_target == 'logfile':
            self.set_visit_message(self.logfile_database_properties.friendly_name)

    def signatures_updated(self, type, item):
        if debug:
            log_gui.debug("signatures_updated() browser: type=%s, item=%s alert_list size=%s", type, item, len(self.alert_list_model))

        def new_siginfo_callback(sigs):
            if debug:
                log_gui.debug("new_siginfo_callback(), type=%s %s", type, str(sigs))

            for siginfo in sigs.signature_list:
                self.insert_siginfo_into_model(siginfo)
            self.update_alert_view()

        database = self.get_visit_database()

        if type == 'add' or type == 'modify':
            async_rpc = database.query_alerts(item)
            async_rpc.add_callback(new_siginfo_callback)
        elif type == 'delete':
            iter = self.find_local_id_in_model(item)
            if iter is not None:
                self.alert_list_model.remove(iter)
        else:
            raise ProgramError(ERR_UNKNOWN_VALUE, "signatures_updated: type = %s not recognized" % type)
            
    def on_signatures_updated(self, alert_client, type, item):
        self.signatures_updated(type, item)

    def validate_data(self):
        timestamp = TimeStamp()
        for siginfo in self.sigs.signature_list:
            if siginfo.first_seen_date is None:
                siginfo.first_seen_date = timestamp
            if siginfo.last_seen_date is None:
                siginfo.last_seen_date = timestamp

    def load_data(self):
        self.clear_data()

        database = self.get_visit_database()
        if debug:
            log_gui.debug("load_data(): visit_target=%s database=%s", self.visit_target, database)
        if database is not None:
            criteria = '*'
            self.set_status_message(_('loading data ...'))
            #self.statusbar_widget.progress.set_fraction(.5)
            self.load_in_progress = True
            self.statusbar_widget.progress.pulse()
            gobject.timeout_add(100, self.progress_pulse)
            self.statusbar_widget.progress.set_text(_("Server Load"))
            async_rpc = database.query_alerts(criteria)
            async_rpc.add_callback(self.query_alerts_callback)
            async_rpc.add_errback(self.query_alerts_error)


    def query_alerts_callback(self, sigs):
        if debug:
            log_gui.debug("query_alerts_callback():")
        self.sigs = sigs
        self.validate_data()
        self.init_gui_data()
        self.set_status_message(_('loading data done'), 15)
        self.load_in_progress = False
        self.statusbar_widget.progress.set_text("")
        self.statusbar_widget.progress.set_fraction(0)


    def query_alerts_error(self, method, errno, strerror):
        log_gui.error("query_alerts: [%d] %s", errno, strerror)
        self.set_status_message("_(load data failed: %s", strerror)
        self.load_in_progress = False
        self.statusbar_widget.progress.set_text("")
        self.statusbar_widget.progress.set_fraction(0)


    def clear_data(self):
        self.sigs = SEFaultSignatureSet()
        self.init_gui_data()
        self.clear_alert_view()

    def save_siginfos(self, file, siginfos):
        text = ''
        try:
            fd=open(file, "w")
        except IOError, e:
            raise ProgramError(ERR_FILE_OPEN, detail="%s, %s" % (filepath, e.strerror))

        i = 0
        for siginfo in siginfos:
            text += siginfo.format_text()
            i += 1
            if i < len(siginfos):
                text += "\f\n"      # break between multiple siginfo's

        fd.write(text)
        fd.close()
        
    # --- Event Handlers ---

    def on_copy(self, widget):
        text = gtkhtml2.html_selection_get_text(self.description_view)

        if debug:
            log_gui.debug("on_copy(): text=%s", text)

        if text is not None:
            self.clipboard.set_text(text)
        return True     # return True ==> handled, False ==> propagate

    def on_copy_alert(self, widget):
        text = ''
        siginfos = self.get_selected_siginfos()
        i = 0
        for siginfo in siginfos:
            text += siginfo.format_text()
            i += 1
            if i < len(siginfos):
                text += "\n"            # break between multiple siginfo's
        self.clipboard.set_text(text)
        return True     # return True ==> handled, False ==> propagate
        

    def on_close(self, *args):
        if self.window_delete_hides:
            self.hide()
            return True
        else:
            gtk.main_quit()
        return True     # return True ==> handled, False ==> propagate
            
    def on_connection_state_change(self, alert_client, connection_state):

        if debug:
            log_gui.debug("on_connection_state_change(): state=%s", connection_state)

        self.update_connection_state(connection_state)

    def on_connection_pending_retry(self, alert_client, seconds_pending):
        if seconds_pending <= 0:
            self.set_connection_status(False, _('retrying connection'))
        else:
            minutes = seconds_pending / 60
            seconds = seconds_pending % 60
            if minutes:
                pending = _("%d minutes %d seconds") % (minutes, seconds)
            else:
                pending = _("%d seconds") % (seconds)
                self.set_connection_status(False, _('retrying connection in %s' % pending))
            

    def on_about(self, widget):
        dlg = self.wTree.get_widget("aboutdialog")
        dlg.run()
        dlg.hide()
        return True     # return True ==> handled, False ==> propagate

    def on_save_as(self, widget):
        dialog = gtk.FileChooserDialog("Save..",
                                       None,
                                       gtk.FILE_CHOOSER_ACTION_SAVE,
                                       (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                        gtk.STOCK_SAVE, gtk.RESPONSE_OK))
        dialog.set_default_response(gtk.RESPONSE_OK)

        if self.default_save_folder is not None:
            dialog.set_current_folder(self.default_save_folder)
        if self.default_save_filename is not None:
            dialog.set_current_name(self.default_save_filename)

        filter = gtk.FileFilter()
        filter.set_name("All files")
        filter.add_pattern("*")
        dialog.add_filter(filter)

        response = dialog.run()
        if response == gtk.RESPONSE_OK:
            self.default_save_folder = dialog.get_current_folder()
            self.default_save_filename = os.path.basename(dialog.get_filename())
            try:
                self.save_siginfos(dialog.get_filename(), self.get_selected_siginfos())
            except ProgramError, e:
                self.display_error(e.strerror)

        dialog.destroy()
        return True     # return True ==> handled, False ==> propagate

    def on_open_logfile(self, widget):
        title = _("Open logfile for scanning")
        options = ''
        dialog = FileChooserDialog()
        filename = dialog.get_filename(title)
        if filename is None:
            return
        if debug:
            log_gui.debug("on_open_logfile: filename=%s", filename)

        try:
            self.logfile_thread = LogfileThread(self, filename)
            self.logfile_thread.run()
        except ProgramError, e:
            self.display_error(e.strerror)
        return True     # return True ==> handled, False ==> propagate

    def on_print(self, widget):

        def draw_page(operation, context, page_number, siginfos):
            try:
                siginfo = siginfos[page_number]
            except IndexError:
                return
            text = siginfo.format_text()

            cr = context.get_cairo_context()
            layout = context.create_pango_layout()
            layout.set_font_description(pango.FontDescription('monospace 10'))
            layout.set_text(text)
            cr.show_layout(layout)

        siginfos = self.get_selected_siginfos()

        dialog = gtk.PrintOperation()
        if self.print_settings != None:
            dialog.set_print_settings(self.print_settings)
        dialog.connect("draw_page", draw_page, siginfos)
        dialog.set_n_pages(len(siginfos))

        res = dialog.run(gtk.PRINT_OPERATION_ACTION_PRINT_DIALOG)
        if res == gtk.PRINT_OPERATION_RESULT_ERROR:
            error_dialog = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT,
                                             gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE,
                                             "Error printing file:\n")
            error_dialog.connect("response", lambda w,id: w.destroy())
            error_dialog.show()
        elif res == gtk.PRINT_OPERATION_RESULT_APPLY:
            self.print_settings = dialog.get_print_settings()
        return True     # return True ==> handled, False ==> propagate
      
    def on_hide_deleted_toggle(self, widget):
        self.hide_deleted = widget.get_active()
        if debug:
            log_gui.debug("on_hide_deleted_toggle() hide_deleted=%s", self.hide_deleted)
        self.alert_list_model_filter.refilter()
        return True     # return True ==> handled, False ==> propagate

    def on_view_audit(self, widget):
        if debug:
            log_gui.debug("on_view_audit()")
        self.update_visit('audit')
        return True     # return True ==> handled, False ==> propagate

    def on_view_logfile(self, widget):
        if debug:
            log_gui.debug("on_view_logfile()")
        self.update_visit('logfile')
        return True     # return True ==> handled, False ==> propagate

    def on_selection_changed(self, selection):
        if debug:
            log_gui.debug("on_selection_changed(): selection=%s", selection)

        gobject.idle_add(self.update_alert_view)
        return True     # return True ==> handled, False ==> propagate

    def on_filter_toggle(self, cell, path, model):
        iter = model.get_iter_from_string(path)
        siginfo = model.get_value(iter, PYOBJECT_COLUMN)
        user_data = siginfo.get_user_data(self.username)

        if user_data.filter.filter_type == FILTER_NEVER:
            user_data.filter.filter_type = FILTER_ALWAYS
        else:
            user_data.filter.filter_type = FILTER_NEVER
        database = self.get_visit_database()
        database.set_filter(siginfo.sig, self.username, user_data.filter.filter_type, '')
        return True     # return True ==> handled, False ==> propagate

    def on_delete(self, widget):
        database = self.get_visit_database()
        for siginfo in self.get_selected_siginfos():
            self.mark_delete(database, siginfo, True)
        return True     # return True ==> handled, False ==> propagate

    def on_expunge(self, widget):
        database = self.get_visit_database()
        for siginfo in self.get_siginfos():
            user_data = siginfo.get_user_data(self.username)
            if user_data.delete_flag:
                database.delete_signature(siginfo.sig)
        return True     # return True ==> handled, False ==> propagate

    def on_alert_list_button_press(self, treeview, event):
        if debug:
            log_gui.debug("on_treeview_button_press_event(): button=%s", event.button)
        if event.button == 3:
            x = int(event.x)
            y = int(event.y)
            time = event.time
            path_info = treeview.get_path_at_pos(x, y)
            if path_info is not None:
                path, col, cell_x, cell_y = path_info
                if debug:
                    log_gui.debug("on_treeview_button_press_event(): path=%s", path)
                #treeview.grab_focus()
                #treeview.set_cursor(path)
                self.alert_list_right_click_menu_widget.popup( None, None, None, event.button, time)
        return False     # return True ==> handled, False ==> propagate

    def on_undelete(self, widget):
        database = self.get_visit_database()
        for siginfo in self.get_selected_siginfos():
            self.mark_delete(database, siginfo, False)
        return True     # return True ==> handled, False ==> propagate

    def on_select_all(self, widget):
        self.alert_list_selection.select_all()
        return True     # return True ==> handled, False ==> propagate
        
    def on_select_none(self, widget):
        self.alert_list_selection.unselect_all()
        return True     # return True ==> handled, False ==> propagate
        
    # --- Interaction Dialogs ---

    def display_error(self, message):
        self.set_status_message(message)
        return display_error(message)
    
    # --- Status Notification ---

    def update_connection_state(self, connection_state):

        if debug:
            log_gui.debug("on_connection_state_change(): state=%s", connection_state)

        self.connection_state = connection_state

        if connection_state == 'closed':
            self.clear_data()
            self.set_connection_status(False, _('connection closed'))
        elif connection_state == 'open':
            self.set_connection_status(True, _('connected'))
        elif connection_state == 'retry':
            self.clear_data()
        elif connection_state == 'hup':
            self.set_connection_status(False, _('connection lost'))
        elif connection_state == 'error':
            self.set_connection_status(False, _('connection failed'))
        elif connection_state == 'invalid':
            pass
        elif connection_state == 'timeout':
            self.set_connection_status(False, _('connection timed out'))
        else:
            log_gui.error("on_connection_state_change: unknown state (%s)", state)


    def set_status_message(self, msg, timeout=None):
        self.status_message = msg
        self.statusbar_widget.status_msg.set_text(self.status_message)
        if timeout is not None:
            def clear_msg():
                if self.status_message == msg:
                    self.clear_status_message()
                return False            # do not call again
            gobject.timeout_add(timeout*1000, clear_msg)
            

    def clear_status_message(self):
        self.status_message = ''
        self.statusbar_widget.status_msg.set_text(self.status_message)

    def set_visit_message(self, msg):
        self.visit_message = msg
        self.statusbar_widget.visit_msg.set_text(msg)

    def clear_visit_message(self):
        self.statusbar_widget.visit_msg.set_text('')

    def set_connection_status(self, is_connected, msg=None):
        if is_connected:
            self.statusbar_widget.connect_icon.set_from_stock(gtk.STOCK_CONNECT, gtk.ICON_SIZE_SMALL_TOOLBAR)
        else:
            self.statusbar_widget.connect_icon.set_from_stock(gtk.STOCK_DISCONNECT, gtk.ICON_SIZE_SMALL_TOOLBAR)
            self.clear_visit_message()
            
        if msg is not None:
            self.set_status_message(msg, timeout=15)


#-----------------------------------------------------------------------------

# -- Main --

if __name__ == "__main__":
    import getopt

    def usage():
        print '''
        -h help
'''

    try:
        opts, args = getopt.getopt(sys.argv[1:], "h", ["help"])
    except getopt.GetoptError:
        # print help information and exit:
        usage()
        sys.exit(2)

    for o, a in opts:
        if o in ("-h", "--help"):
            usage()
            sys.exit()

    browser_applet = BrowserApplet()
    browser_applet.show()

    gtk.main()
