#
# Copyright (C) 2006 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

# Needed to silence warnings if X display is not present
import warnings
warnings.filterwarnings('ignore', 'could not open display')

from rhpl.translate import _
import pygtk
pygtk.require("2.0")
import gtk, gtk.glade
import gtkhtml2
import pynotify
import gobject 
import gnome
import os
import datetime
import re
import selinux
import traceback

from setroubleshoot.config import cfg
from setroubleshoot.errcode import *
from setroubleshoot.signature import *
from setroubleshoot.util import *
from setroubleshoot.log import *

#------------------------------------------------------------------------------

__all__ = ['map_column_types',
           'Applet',
           'DateTimeDialog',
           'FileChooserDialog',
           'get_display',
           'display_verify',
           'display_error',
           'display_traceback',
           'display_restart',
           'window_state_to_string',
           'parse_window_state',
          ]


#------------------------------------------------------------------------------

map_column_types = {
    'icon'     : gtk.gdk.Pixbuf,
    'string'   : gobject.TYPE_STRING,
    'int'      : gobject.TYPE_INT,
    'pyobject' : gobject.TYPE_PYOBJECT,
    'toggle'   : gobject.TYPE_BOOLEAN
    }

#------------------------------------------------------------------------------

def get_display():
    try:
        dpy = gtk.gdk.Display('')
        dpy_name = dpy.get_name()
        return dpy_name
    except RuntimeError, e:
        return None

    # --- Interaction Dialogs ---

def display_verify(message):
    if get_display() is None:
        return None
    dlg = gtk.MessageDialog(None, 0, gtk.MESSAGE_INFO, gtk.BUTTONS_YES_NO, message)
    dlg.set_position(gtk.WIN_POS_MOUSE)
    dlg.show_all()
    rc = dlg.run()
    dlg.destroy()
    if rc == gtk.RESPONSE_YES:
        return True
    if rc == gtk.RESPONSE_NO:
        return False
    return None
    
def display_error(message):
    if get_display() is None:
        return None
    dlg = gtk.MessageDialog(None, 0, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK, message)
    dlg.set_position(gtk.WIN_POS_CENTER)
    dlg.show_all()
    rc = dlg.run()
    dlg.destroy()
    if rc == gtk.RESPONSE_OK:
        return True
    return None


def display_traceback(who):
    if get_display() is None:
        return None
    stacktrace = traceback.format_exc()
    message = _("Opps, %s hit an error" % who)

    dlg = gtk.MessageDialog(None, 0, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK, message+'\n'+stacktrace)
    dlg.set_position(gtk.WIN_POS_CENTER)
    dlg.show_all()
    rc = dlg.run()
    dlg.destroy()
    if rc == gtk.RESPONSE_OK:
        return True
    return None

def display_restart(seconds, restart_func):
    # This implements a dialog which does not block the main application,
    # note that dialog.run() is not called, we use a response callback
    # for the dialog results, the dialog is forced to always be on top.

    # If no X display we can't do anything
    if get_display() is None:
        return None

    count_down_message = _("Restarting in %d seconds")

    # This is passed in array because nested functions can only modify local and
    # global symbols, but mutable objects such as classes and arrays can be modified
    seconds = [seconds]
    restart_now_id = 1432
    timer_id = None

    def dialog_response_cb(dialog, response_id):
        dialog.destroy()
        if response_id == gtk.RESPONSE_CANCEL:
            if timer_id is not None:
                gobject.source_remove(timer_id)
        if response_id == restart_now_id:
            restart_func()
        
    def set_time(seconds):
        label2.set_text(count_down_message % seconds[0])
        seconds[0] -= 1
        if seconds[0] > 0:
            return True
        dlg.destroy()
        restart_func()
        return False
    
    dlg = gtk.Dialog(None, None, 0, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
    label1 = gtk.Label(_("sealert detected the setroubleshoot package changed versions"))
    label2 = gtk.Label(count_down_message % seconds[0])
    label3 = gtk.Label(_("If this message repeats after restarting sealert you'll need to restart the setroubleshoot service too"))
    dlg.vbox.pack_start(label1, True, True, 0)
    dlg.vbox.pack_start(label2, True, True, 0)
    dlg.vbox.pack_start(label3, True, True, 0)
    dlg.add_button(_("Restart Now"), restart_now_id)
    dlg.connect('response', dialog_response_cb)
    dlg.set_position(gtk.WIN_POS_CENTER)
    dlg.set_keep_above(True)
    dlg.show_all()
    timer_id = gobject.timeout_add(1000, set_time, seconds)
    return True
    
def window_state_to_string(state):
    flags = []
    if state & gtk.gdk.WINDOW_STATE_WITHDRAWN:
        flags.append('hidden')
    if state & gtk.gdk.WINDOW_STATE_ICONIFIED:
        flags.append('iconified')
    if state & gtk.gdk.WINDOW_STATE_MAXIMIZED:
        flags.append('maximized')
    if state & gtk.gdk.WINDOW_STATE_STICKY:
        flags.append('sticky')
    if state & gtk.gdk.WINDOW_STATE_FULLSCREEN:
        flags.append('fullscreen')
    if state & gtk.gdk.WINDOW_STATE_ABOVE:
        flags.append('above')
    if state & gtk.gdk.WINDOW_STATE_BELOW:
        flags.append('below')
    return ','.join(flags)

def parse_window_state(str):
    state = 0

    str = re.sub('[,;: \t]+', ',', str)
    flags = str.lower().split(',')
    for flag in flags:
        if not flag: continue
        if flag == 'hidden':
            state |= gtk.gdk.WINDOW_STATE_WITHDRAWN
        elif flag == 'iconified':
            state |= gtk.gdk.WINDOW_STATE_ICONIFIED
        elif flag == 'maximized':
            state |= gtk.gdk.WINDOW_STATE_MAXIMIZED
        elif flag == 'sticky':
            state |= gtk.gdk.WINDOW_STATE_STICKY
        elif flag == 'fullscreen':
            state |= gtk.gdk.WINDOW_STATE_FULLSCREEN
        elif flag == 'above':
            state |= gtk.gdk.WINDOW_STATE_ABOVE
        elif flag == 'below':
            state |= gtk.gdk.WINDOW_STATE_BELOW
        else:
            print "WARNING: unknown window state = '%s'" % (flag)
        
    return state
        
#------------------------------------------------------------------------------

class Applet:

    def get_widget(self, name):
        widget_name = '%s_widget' % name
        widget = self.wTree.get_widget(name)
        if widget is None:
            raise ProgramError(ERR_CANNOT_CREATE_GUI, "cannot find %s widget, GUI fails" % name)
        
        self.__dict__[widget_name] = widget
        return widget

    def create_htmlview(self, container):
        view = gtkhtml2.View()
        doc = gtkhtml2.Document()
        container.set_hadjustment(view.get_hadjustment())
        container.set_vadjustment(view.get_vadjustment())
        view.set_document(doc)
        container.add(view)

        return (view, doc)

#------------------------------------------------------------------------------

class DateTimeDialog:
    AM = 0
    PM = 1

    def __init__(self):
        program_name = 'setroubleshoot browser'
        program_version = '1.0'
        gnome.init(program_name, program_version)
        self.window_name = "browser"
        self.window_name = "date_time_dialog"
        glade_file = '../gui/setroubleshoot_browser.glade'
        self.wTree = gtk.glade.XML(glade_file, self.window_name)

        self.dialog = self.wTree.get_widget("date_time_dialog")

        self.dt = datetime.datetime.now()

        self.calendar = self.wTree.get_widget("calendar")
        self.hour_spinbutton = self.wTree.get_widget("hour_spinbutton")
        self.minute_spinbutton = self.wTree.get_widget("minute_spinbutton")
        self.am_pm_combobox = self.wTree.get_widget("am_pm_combobox")
        self.cur_datetime_echo = self.wTree.get_widget("cur_datetime_echo")

        self.calendar.connect("month_changed", self.on_update)
        self.calendar.connect("day_selected", self.on_update)
        self.calendar.connect("day_selected_double_click", self.on_update)

        self.hour_spinbutton.connect("value_changed", self.on_update)
        self.minute_spinbutton.connect("value_changed", self.on_update)
        self.am_pm_combobox.connect("changed", self.on_update)

    def on_update(self, widget):
        self.echo_cur_datetime()

    def echo_cur_datetime(self):
        self.get_datetime_from_widgets()
        self.cur_datetime_echo.set_text(self.datetime_echo_string())

    def set_time(self, hour, minute, am_pm):
        date = self.dt.date()
        time = datetime.time(hour, minute)
        self.dt = self.dt.combine(date, time)
        self.update_view()
        return self.dt

    def update_view(self):
        hour = self.dt.hour
        if hour <= 12:
            am_pm = self.AM
        else:
            hour -= 12
            am_pm = self.PM
        minute = self.dt.minute

        year = self.dt.year
        month = self.dt.month - 1
        day = self.dt.day

        self.calendar.select_month(month, year)
        self.calendar.select_day(day)

        self.hour_spinbutton.set_value(hour)
        self.minute_spinbutton.set_value(minute)
        self.am_pm_combobox.set_active(am_pm)
        self.echo_cur_datetime()

    def datetime_echo_string(self):
        if self.am_pm_combobox.get_active() == self.AM:
            am_pm = ' AM'
        else:
            am_pm = ' PM'
        str = self.dt.strftime('%a %b %e %Y %l:%M') + am_pm
        return str

    def get_datetime_from_widgets(self):
        hour = self.hour_spinbutton.get_value_as_int()
        minute = self.minute_spinbutton.get_value_as_int()
        am_pm = self.am_pm_combobox.get_active()
        year, month, day = self.calendar.get_date()
        #print "year=%s month=%s day=%s hour=%s minute=%s" % (year, month, day, hour, minute)
        # calendar widget 1 based, datetime object 0 based
        month += 1
        self.dt = datetime.datetime(year, month, day, hour, minute)
        # adjust for am/pm
        if am_pm == self.PM:
            # pretend 12am == 12pm on the same day
            if hour != 12:
                self.dt += datetime.timedelta(hours=12)
        return self.dt

    def get_datetime(self):
        return self.get_datetime_from_widgets()

    def dialog_get_datetime(self):
        result = None
        self.set_time(12, 0, self.PM)
	if self.dialog.run() == gtk.RESPONSE_OK:
            result = self.get_datetime_from_widgets()
	self.dialog.destroy()
        return result


#------------------------------------------------------------------------------
class FileChooserDialog:
    def __init__(self):
        pass
    
    def get_filename(self, title):
	result = None
	file_open = gtk.FileChooserDialog(title, action=gtk.FILE_CHOOSER_ACTION_OPEN,
            buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OPEN, gtk.RESPONSE_OK))

	if file_open.run() == gtk.RESPONSE_OK:
		result = file_open.get_filename()
	file_open.destroy()
        return result
        
#------------------------------------------------------------------------------
